/********************************************************************************
* Copyright (c) Erik Kunze 1995 - 1999
*
* Permission to use, distribute, and sell this software and its documentation
* for any purpose is hereby granted without fee, provided that the above
* copyright notice appear in all copies and that both that copyright notice and
* this permission notice appear in supporting documentation, and that the name
* of the copyright holder not be used in advertising or publicity pertaining to
* distribution of the software without specific, written prior permission.  The
* copyright holder makes no representations about the suitability of this
* software for any purpose.  It is provided "as is" without express or implied
* warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
*
* THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
* INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
* EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
* CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
* DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
* TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
* PERFORMANCE OF THIS SOFTWARE.
*
* Author: Erik Kunze
*******************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#ifndef lint
static char rcsid[] = "$Id: loadsave.c,v 4.36 1999/07/26 19:07:18 erik Rel $";
#endif
#ifndef REGISTERED
#include <X11/keysym.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <setjmp.h>
#include <assert.h>
#include "z80.h"
#include "debug.h"
#include "resource.h"
#include "mem.h"
#include "machine.h"
#include "util.h"
#include "dialog.h"
#include "tap.h"
#include "tzx.h"
#ifdef REGISTERED
#include "voc.h"
#endif
#include "vtp.h"
#include "loadsave.h"
static void tpRecordDestroy(void *);
static uns8 tpDummyGetNextBit(void);
static void tpCatchFault(int);
unsigned short TpState;
uns8 *TpInBegin;
uns8 *TpInEnd;
Llist *TpInBlockList = NULL;
unsigned int TpInBlockCount = 0;
unsigned int TpInBlockCurrent = 0;
uns8 (*TpGetNextBit)(void) = tpDummyGetNextBit;
static char *tpInFn = NULL;
static char *tpOutFn = NULL;
static FILE *tpOutFp;
static jmp_buf tpEnv;
void
TpInit(void)
{
TpState &= ~(TS_QUICKLOAD | TS_FASTLOAD | TS_NOISE | TS_AUTOSTOP);
#ifdef SPEAKER_AUDIO
TpState |= GETCFG(tapeNoise) ? TS_NOISE : 0;
#endif
#ifdef REGISTERED
TpState |= (GETCFG(tapeQuickload) ? TS_QUICKLOAD : 0)
| (GETCFG(tapeFastload) ? TS_FASTLOAD : 0)
| (GETCFG(tapeAutostop) ? TS_AUTOSTOP : 0);
#endif
}
#ifndef NO_ED_TRAPS
void
TpLoadBlock(void)
{
#ifdef DEBUG
if (GETCFG(debug) & D_TAPE)
{
Msg(M_DEBUG, "%s request: id = %02x, addr = %04X, length = %5d",
F1 & C_FLAG ? "load" : "verify", A1, IX, DE);
}
#endif
if (!(TpState & (TS_TAPIN | TS_TZXIN | TS_VOCIN)) || TpState & TS_ENDIN)
{
#ifdef REGISTERED
(void)TpInsertTape("Load Tape", TO_READ | TO_TAP | TO_TZX | TO_VOC);
#else
(void)TpInsertTape("Load Tape", TO_READ | TO_TAP | TO_TZX);
#endif
}
if ((TpState & (TS_TAPIN | TS_TZXIN | TO_VOC))
&& !(TpState & (TS_PLAY | TS_ENDIN | TS_ERRIN)))
{
TpStartPlayback();
}
if 	((TpState & (TS_TAPIN | TS_TZXIN))
&& (TpState & (TS_PLAY | TS_QUICKLOAD)) == (TS_PLAY | TS_QUICKLOAD))
{
if (!TzxLoadBlock())
{
return;
}
}
TSTATES += 11 - 8;
DEC_R();
A = Z80_IN((A << 8) | 0xfe);
}
#endif
#ifndef NO_ED_TRAPS
void
TpSaveBlock(void)
{
#ifdef DEBUG
if (GETCFG(debug) & D_TAPE)
{
Msg(M_DEBUG, "%stape saving: addr = %04X, length = %04X",
A ? "headerless " : "",	IX, DE);
}
#endif
if (!(TpState & TS_TAPOUT))
{
(void)TpInsertTape("Save Tape", TO_WRITE | TO_TAP);
}
if (TpState & TS_TAPOUT)
{
TapSaveBlock(tpOutFp);
return;
}
TSTATES += 10 - 8;
DEC_R();
HL = 0x053f;
PC++;
}
#endif
int
TpInsertTape(const char *title, int operation)
{
int types[4];
char *fname;
int readonly;
size_t size;
struct sigaction newact, oldact;
int i = 0, ret = -1;
if (operation & TO_TAP)
{
types[i++] = FT_TAP;
}
if (operation & TO_TZX)
{
types[i++] = FT_TZX;
}
#ifdef REGISTERED
if (operation & TO_VOC)
{
assert(operation & TO_READ);
types[i++] = FT_VOC;
}
#endif
types[i] = FT_END;
if (operation & TO_READ)
{
if (!(fname = FileSelector(title, 1, types)))
{
return -1;
}
TpEjectTape(TO_READ);
readonly = 1;
TpInBegin = (uns8 *)Mmap(fname, 1, &readonly, &size);
if (TpInBegin != (uns8 *)-1)
{
TpInEnd = TpInBegin + size;
tpInFn = fname;
newact.sa_handler = tpCatchFault;
(void)sigemptyset(&newact.sa_mask);
#ifdef SA_RESTART
newact.sa_flags = SA_RESTART;
#else
newact.sa_flags = 0;
#endif
(void)sigaction(SIGSEGV, &newact, &oldact);
if (!setjmp(tpEnv))
{
switch (GetFileType(fname))
{
case FT_TAP:
TpState |= TS_TAPIN;
ret = TapNewFile();
break;
case FT_TZX:
TpState |= TS_TZXIN;
ret = TzxNewFile();
break;
#ifdef REGISTERED
case FT_VOC:
TpState |= TS_VOCIN;
ret = VocNewFile();
break;
#endif
}
}
else
{
Msg(M_ERR, "corrupted input file");
ret = -1;
}
(void)sigaction(SIGSEGV, &oldact, NULL);
if (ret != 0)
{
TpState &= ~(TS_TAPIN | TS_TZXIN | TS_VOCIN);
Munmap((void *)TpInBegin, size);
free(tpInFn);
tpInFn = NULL;
DestroyList(TpInBlockList, tpRecordDestroy);
TpInBlockList = NULL;
TpInBlockCount = 0;
TpInBlockCurrent = 0;
}
}
else
{
free(fname);
}
}
else
{
if ((fname = FileSelector(title, 0, types)))
{
TpEjectTape(TO_WRITE);
if (!(tpOutFp = Fopen(fname, "wb")))
{
Msg(M_PERR, "couldn't open <%s> for writing", fname);
free(fname);
}
else
{
TpState |= TS_TAPOUT;
tpOutFn = fname;
VtpNewOutputFile();
ret = 0;
}
}
}
return ret;
}
void
TpEjectTape(int operation)
{
if (operation & TO_READ)
{
if (TpState & (TS_TAPIN | TS_TZXIN | TS_VOCIN))
{
TpGetNextBit = tpDummyGetNextBit;
TpState &= ~(TS_TAPIN | TS_TZXIN | TS_VOCIN | TS_PLAY | TS_ERRIN |
TS_ENDIN);
Munmap((void *)TpInBegin, (size_t)(TpInEnd - TpInBegin));
free(tpInFn);
tpInFn = NULL;
DestroyList(TpInBlockList, tpRecordDestroy);
TpInBlockList = NULL;
TpInBlockCount = 0;
TpInBlockCurrent = 0;
}
}
else
{
if (TpState & TS_TAPOUT)
{
TpState &= ~(TS_TAPOUT | TS_ERROUT | TS_ENDOUT);
(void)fclose(tpOutFp);
free(tpOutFn);
tpOutFn = NULL;
}
}
}
static void
tpRecordDestroy(void *entry)
{
if (((TapeRecord *)entry)->str)
{
free(((TapeRecord *)entry)->str);
}
}
char *
TpTapeName(int operation)
{
return (operation == TO_READ ? tpInFn : tpOutFn);
}
void
TpStartPlayback(void)
{
assert(TpState & (TS_TAPIN | TS_TZXIN | TS_VOCIN)
&& !(TpState & (TS_PLAY | TS_ERRIN | TS_ENDIN)));
TpState |= TS_PLAY;
switch (TpState & (TS_TAPIN | TS_TZXIN | TS_VOCIN))
{
case TS_TAPIN:
case TS_TZXIN:
TpGetNextBit = TzxGetNextBit;
TzxStartPlaying();
break;
#ifdef REGISTERED
case TS_VOCIN:
TpGetNextBit = VocGetNextBit;
VocStartPlaying();
break;
#endif
}
VtpUpdate();
}
void
TpStopPlayback(void)
{
assert(TpState & TS_PLAY);
TpState &= ~TS_PLAY;
TpGetNextBit = tpDummyGetNextBit;
VtpUpdate();
}
static uns8
tpDummyGetNextBit(void)
{
return 0;
}
static void
tpCatchFault(int signo)
{
longjmp(tpEnv, -1);
}
