/********************************************************************************
* Copyright (c) Erik Kunze 1995 - 1999
*
* Permission to use, distribute, and sell this software and its documentation
* for any purpose is hereby granted without fee, provided that the above
* copyright notice appear in all copies and that both that copyright notice and
* this permission notice appear in supporting documentation, and that the name
* of the copyright holder not be used in advertising or publicity pertaining to
* distribution of the software without specific, written prior permission.  The
* copyright holder makes no representations about the suitability of this
* software for any purpose.  It is provided "as is" without express or implied
* warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
*
* THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
* INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
* EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
* CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
* DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
* TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
* PERFORMANCE OF THIS SOFTWARE.
*
* Author: Erik Kunze
*******************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#ifdef XZX_PENTAGON
#ifndef lint
static char rcsid[] = "$Id: pentagon.c,v 1.14 1999/05/22 14:48:22 erik Rel $";
#endif
#include "z80.h"
#include "debug.h"
#include "resource.h"
#include "mem.h"
#include "io.h"
#include "machine.h"
#include "util.h"
#ifdef AUDIO
#include "audio.h"
#endif
#include "emul.h"
#ifdef JOY
#include "joystick.h"
#endif
#include "keyboard.h"
#ifdef XZX_KMOUSE
#include "kmouse.h"
#endif
#include "loadsave.h"
#include "screen.h"
#include "tables.h"
#include "trdos.h"
#include "pentagon.h"
#ifdef DEBUG
#define DEB(x)			{ if (GETCFG(debug) & D_IO) { x } }
#else
#define DEB(x)
#endif
static void pentReset(void);
static void pentNmi(void);
static uns8 pentReadOpcode(uns16);
static uns8 pentReadMem(uns16);
static void pentWriteMem(uns16, uns8);
static void pentInitPages(void);
static void pentPageIn(int, int);
static inline void pentOutBankm128(uns8);
static void pentOutPort(uns16, uns8);
static uns8 pentInPort(uns16);
void
PentInit(void)
{
MachineSettings *msp;
int checksum, romtype, i, j, k;
msp = &Machines[SP_PENTAGON];
msp->name = "Pentagon";
msp->basicRom = PENTROM1;
msp->romType = ROM_UNKNOWN;
msp->tvLines = 320;
msp->tstatesPerLine = 224;
msp->tstatesPerFrame = 71680;
msp->intLine = 304;
msp->topBorder = 64;
msp->ulaDelay = Malloc(msp->tstatesPerFrame, "PentInit");
msp->reset = pentReset;
msp->nmi = pentNmi;
msp->readOpcode = pentReadOpcode;
msp->readMem = pentReadMem;
msp->writeMem = pentWriteMem;
msp->inPort = pentInPort;
msp->outPort = pentOutPort;
msp->initPages = pentInitPages;
msp->pageIn = pentPageIn;
if (!LoadRomImage(GETCFG(pentRom), PENTROM0, 3))
{
checksum = GetRomChecksum(msp->basicRom, XZX_PAGESIZE);
switch (checksum)
{
case 44809:
romtype = ROM_PENTAGON;
break;
default:
Msg(M_WARN, "unknown Pentagon ROM image - checksum %u",
checksum);
romtype = ROM_UNKNOWN;
break;
}
msp->romType = romtype;
#ifndef NO_ED_TRAPS
if (romtype != ROM_UNKNOWN)
{
RealMemory[msp->basicRom][0x0562] = 0xed;
RealMemory[msp->basicRom][0x0563] = LOAD_BPT;
RealMemory[msp->basicRom][0x04c2] = 0xed;
RealMemory[msp->basicRom][0x04c3] = SAVE_BPT;
#ifdef XBELL_AUDIO
RealMemory[msp->basicRom][0x03b5] = 0xed;
RealMemory[msp->basicRom][0x03b6] = BELL_BPT;
#endif
}
#endif
}
for (i = 0; i < 14335; i++)
{
msp->ulaDelay[i] = 0x80;
}
for (j = 192; j; j --)
{
for (k = 128; k; k--)
{
msp->ulaDelay[i++] = 0;
}
for (k = 96; k; k--)
{
msp->ulaDelay[i++] = 0x80;
}
}
for (; i < msp->tstatesPerFrame; i++)
{
msp->ulaDelay[i] = 0x80;
}
}
static void
pentReset(void)
{
Z80_Reset();
Machine->initPages();
Last0x7FFD = 0;
ScreenSelect = RAM5;
KbdReset();
(void)memset(InPorts, 0x00, sizeof(InPorts));
(void)memset(OutPorts, 0x00, sizeof(OutPorts));
Last0xFFFD = 0;
(void)memset(Psg, 0x00, sizeof(Psg));
#ifdef AYCHIP_AUDIO
AyReset();
#endif
}
static void
pentNmi(void)
{
Z80_NMI();
Machine->pageIn(0, PENTROM2);
}
static uns8
pentReadOpcode(uns16 addr)
{
uns8 op;
if (addr >= 0x4000)
{
if (RPAGE(0) == PENTROM2)
{
Machine->pageIn(0, Machine->basicRom);
}
}
else if ((addr & 0xff00) == 0x3d00)
{
if (RPAGE(0) == Machine->basicRom)
{
Machine->pageIn(0, PENTROM2);
}
}
op = MEM(addr);
return op;
}
static uns8
pentReadMem(uns16 addr)
{
uns8 byte;
byte = MEM(addr);
return byte;
}
static void
pentWriteMem(uns16 addr, uns8 byte)
{
int page = VPAGE(addr);
if (RealPageNumbers[page] >= RAM0)
{
TheMemory[page][addr & 0x1fff] = byte;
if (RealPageNumbers[page] == ScreenSelect)
{
if ((addr &= 0x3fff) < (uns16)PIXEL_LENGTH)
{
ScreenWrite(addr, byte);
}
else if (addr < (uns16)(PIXEL_LENGTH + ATTR_LENGTH))
{
AttrWrite(addr, byte);
}
}
}
}
static void
pentInitPages(void)
{
Machine->pageIn(0, PENTROM0);
Machine->pageIn(1, RAM5);
Machine->pageIn(2, RAM2);
Machine->pageIn(3, RAM0);
}
static void
pentPageIn(int z80page, int vpage)
{
int i = z80page << 1;
TheMemory[i] = RealMemory[vpage];
RealPageNumbers[i] = vpage;
i++;
TheMemory[i] = RealMemory[vpage] + XZX_PAGESIZE / 2;
RealPageNumbers[i] = vpage;
}
static inline void
pentOutBankm128(uns8 val)
{
DEB(Msg(M_DEBUG,
"bankm: RAM select %d, ROM select %d, screen %d, paging %sabled",
val & B_SELRAM,
(val & B_SELROM) >> 4, (val & B_SELSCREEN) >> 3,
PAGING_DISABLED(Last0x7FFD) ? "dis" : "en"););
if (Last0x7FFD == val || PAGING_DISABLED(Last0x7FFD))
{
return;
}
ScreenSelect = val & B_SELSCREEN ? RAM7 : RAM5;
if ((val & B_SELSCREEN) != (Last0x7FFD & B_SELSCREEN))
{
ForceScreenRefresh();
}
Last0x7FFD = val;
Machine->pageIn(3, RAM0 + (val & B_SELRAM));
Machine->pageIn(0, PENTROM0 + ((val & B_SELROM) >> 4));
}
static void
pentOutPort(uns16 port, uns8 val)
{
DEB(Msg(M_DEBUG, "writing <%02x> to port <%04x>", val, port););
if (RPAGE(0) == PENTROM2)
{
TrdosOutByte(port, val);
}
if ((port & B_ULA) == (P_ULA & B_ULA))
{
if (LastBorderColor != (val & B_BORDER))
{
LastBorderColor = val & B_BORDER;
SetBorderColor(LastBorderColor);
}
#ifdef SPEAKER_AUDIO
SpeakerOutByte(val);
#endif
#ifdef EMULATE_ULA_DELAY
#endif
}
if ((port & B_128) == (P_128 & B_128))
{
if ((port & B_SNDCHIP) == (P_SNDCONTROL & B_SNDCHIP))
{
if ((port & B_SNDCONTROL) == (P_SNDCONTROL & B_SNDCONTROL))
{
Last0xFFFD = val;
}
else
{
if (Last0xFFFD < NELEM(Psg))
{
Psg[Last0xFFFD] = val;
if (Last0xFFFD == 14)
{
DEB(Msg(M_DEBUG, "writing AY register 14"););
}
#ifdef AYCHIP_AUDIO
else if (Last0xFFFD < 14)
{
AyOutByte(Last0xFFFD, val);
}
#endif
}
else
{
DEB(Msg(M_DEBUG, "writing AY register %d", Last0xFFFD););
}
}
}
else
{
pentOutBankm128(val);
}
}
OutPorts[LOWBYTE(port)] = val;
}
static uns8
pentInPort(uns16 port)
{
uns8 res;
if (RPAGE(0) == PENTROM2
&& (port & 0x001f) == 0x001f
&& ((port & 0x0080) == 0x0000 || (port & 0x00f0) == 0x00f0))
{
res = TrdosInByte(port);
goto quit;
}
#ifdef XZX_KMOUSE
INBYTE_KMOUSE(port, res, quit);
#endif
if ((port & B_KEMPSTON) == (P_KEMPSTON & B_KEMPSTON))
{
#ifdef JOY
if (GETCFG(joyActive) > 0)
{
InPorts[P_KEMPSTON] = (uns8)JoyRead();
}
#endif
res = InPorts[P_KEMPSTON] &
(B_RIGHT | B_LEFT | B_DOWN | B_UP | B_FIRE);
goto quit;
}
if ((port & B_ULA) == (P_ULA & B_ULA))
{
res = 0xbf;
if (!(port & 0x8000))
{
res &= KeyPorts[7];
}
if (!(port & 0x4000))
{
res &= KeyPorts[6];
}
if (!(port & 0x2000))
{
res &= KeyPorts[5];
}
if (!(port & 0x1000))
{
res &= KeyPorts[4];
}
if (!(port & 0x0800))
{
res &= KeyPorts[3];
}
if (!(port & 0x0400))
{
res &= KeyPorts[2];
}
if (!(port & 0x0200))
{
res &= KeyPorts[1];
}
if (!(port & 0x0100))
{
res &= KeyPorts[0];
}
res ^= TpGetNextBit();
#ifdef EMULATE_ULA_DELAY
#endif
goto quit;
}
if ((port & B_128) == (P_128 & B_128))
{
if ((port & B_SNDCHIP) == (P_SNDCONTROL & B_SNDCHIP))
{
if ((port & B_SNDCONTROL) == (P_SNDCONTROL & B_SNDCONTROL))
{
if (Last0xFFFD == 14)
{
DEB(Msg(M_DEBUG, "reading AY register 14"););
res = Psg[7] & 0x40 ? (Psg[14] & 0x1f) | 0x20 : 0x3f;
}
else if (Last0xFFFD < 14)
{
res = (Psg[Last0xFFFD] & PsgMask[Last0xFFFD]);
}
else if (Last0xFFFD == 15)
{
res = (Psg[7] & 0x80 ? Psg[15] : 0xff);
}
else
{
res = 0xff;
}
}
else
{
res = 0xff;
}
goto quit;
}
}
if (!(Machine->ulaDelay[TSTATES % Machine->tstatesPerFrame] & 0x80))
{
unsigned int tstatesInLine = TSTATES % Machine->tstatesPerLine;
unsigned int line = Vline - Machine->topBorder;
res = RealMemory[ScreenSelect][(tstatesInLine & 0x02 ?
Line2Attr[line] : Line2Pix[line]) +
(tstatesInLine >> 2)];
}
else
{
res = 0xff;
}
quit:
DEB(Msg(M_DEBUG, "reading <%02x> from port <%04x>", res, port););
return res;
}
#endif
