/********************************************************************************
* Copyright (c) Erik Kunze 1995 - 1999
*
* Permission to use, distribute, and sell this software and its documentation
* for any purpose is hereby granted without fee, provided that the above
* copyright notice appear in all copies and that both that copyright notice and
* this permission notice appear in supporting documentation, and that the name
* of the copyright holder not be used in advertising or publicity pertaining to
* distribution of the software without specific, written prior permission.  The
* copyright holder makes no representations about the suitability of this
* software for any purpose.  It is provided "as is" without express or implied
* warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
*
* THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
* INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
* EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
* CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
* DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
* TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
* PERFORMANCE OF THIS SOFTWARE.
*
* Author: Erik Kunze
*******************************************************************************/
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#ifndef lint
static char rcsid[] = "$Id: spec128.c,v 1.2 1999/07/26 18:54:12 erik Rel $";
#endif
#include "z80.h"
#include "debug.h"
#include "resource.h"
#include "mem.h"
#include "io.h"
#include "machine.h"
#include "util.h"
#ifdef AUDIO
#include "audio.h"
#endif
#include "emul.h"
#ifdef XZX_IF1
#include "if1.h"
#endif
#ifdef JOY
#include "joystick.h"
#endif
#include "keyboard.h"
#include "loadsave.h"
#ifdef XZX_MF
#include "multiface.h"
#endif
#ifdef XZX_MGT
#include "mgt.h"
#endif
#ifdef PRINTER
#include "printer.h"
#endif
#include "screen.h"
#include "tables.h"
#ifdef XZX_TRDOS
#include "trdos.h"
#endif
#include "spec128.h"
#ifdef DEBUG
#define DEB(x)			{ if (GETCFG(debug) & D_IO) { x } }
#else
#define DEB(x)
#endif
static void spec128Reset(void);
static void spec128Nmi(void);
static uns8 spec128ReadOpcode(uns16);
static uns8 spec128ReadMem(uns16);
static void spec128WriteMem(uns16, uns8);
static void spec128InitPages(void);
static void spec128PageIn(int, int);
static inline void spec128OutBankm128(uns8);
static void spec128OutPort(uns16, uns8);
static uns8 spec128InPort(uns16);
void
Spec128Init(void)
{
MachineSettings *msp;
int checksum, romtype, i, j, k;
msp = &Machines[SP_128];
msp->name = "Spectrum 128K";
msp->basicRom = SP128ROM1;
msp->romType = ROM_UNKNOWN;
msp->tvLines = 311;
msp->tstatesPerLine = 228;
msp->tstatesPerFrame = 70908;
msp->intLine = 0;
msp->topBorder = 63;
msp->ulaDelay = Malloc(msp->tstatesPerFrame, "Spec128Init");
msp->reset = spec128Reset;
msp->nmi = spec128Nmi;
msp->readOpcode = spec128ReadOpcode;
msp->readMem = spec128ReadMem;
msp->writeMem = spec128WriteMem;
msp->inPort = spec128InPort;
msp->outPort = spec128OutPort;
msp->initPages = spec128InitPages;
msp->pageIn = spec128PageIn;
if (!LoadRomImage(GETCFG(sp128Rom), SP128ROM0, 2))
{
checksum = GetRomChecksum(msp->basicRom, XZX_PAGESIZE);
switch (checksum)
{
case 44809:
romtype = ROM_128;
break;
case 32991:
msp->name = "Spectrum +2";
romtype = ROM_PLUS2;
break;
default:
Msg(M_WARN, "unknown 128K ROM image - checksum %u", checksum);
romtype = ROM_UNKNOWN;
break;
}
msp->romType = romtype;
#ifndef NO_ED_TRAPS
if (romtype != ROM_UNKNOWN)
{
RealMemory[msp->basicRom][0x0562] = 0xed;
RealMemory[msp->basicRom][0x0563] = LOAD_BPT;
RealMemory[msp->basicRom][0x04c2] = 0xed;
RealMemory[msp->basicRom][0x04c3] = SAVE_BPT;
#ifdef XBELL_AUDIO
RealMemory[msp->basicRom][0x03b5] = 0xed;
RealMemory[msp->basicRom][0x03b6] = BELL_BPT;
#endif
}
#endif
}
for (i = 0; i < 14361; i++)
{
msp->ulaDelay[i] = 0x80;
}
for (j = 192; j; j --)
{
for (k = 128; k; k--)
{
msp->ulaDelay[i++] = 4;
}
for (k = 100; k; k--)
{
msp->ulaDelay[i++] = 0x80;
}
}
for (; i < msp->tstatesPerFrame; i++)
{
msp->ulaDelay[i] = 0x80;
}
}
static void
spec128Reset(void)
{
Z80_Reset();
Machine->initPages();
Last0x7FFD = 0;
ScreenSelect = RAM5;
KbdReset();
(void)memset(InPorts, 0x00, sizeof(InPorts));
(void)memset(OutPorts, 0x00, sizeof(OutPorts));
Last0xFFFD = 0;
(void)memset(Psg, 0x00, sizeof(Psg));
#ifdef AYCHIP_AUDIO
AyReset();
#endif
#ifdef XZX_TRDOS
TrdosReset();
#endif
#ifdef XZX_MGT
MgtReset();
#endif
}
static void
spec128Nmi(void)
{
Z80_NMI();
#ifdef XZX_MF
if (GETCFG(mfActive))
{
Machine->pageIn(0, MF128ROM);
}
#endif
#ifdef XZX_TRDOS
if (GETCFG(trdosActive) > 0)
{
Machine->pageIn(0, TRDOSROM);
}
#endif
#ifdef XZX_MGT
if (GETCFG(mgtActive) > 0)
{
Machine->pageIn(0, MGTROM);
}
#endif
}
static uns8
spec128ReadOpcode(uns16 addr)
{
uns8 op;
#ifdef XZX_IF1
CHECK_AND_PAGEIN_IF1(addr);
#endif
#ifdef XZX_TRDOS
CHECK_TRDOS(addr);
#endif
#ifdef XZX_MGT
CHECK_MGT(addr);
#endif
op = MEM(addr);
#ifdef EMULATE_ULA_DELAY
if (RPAGE(addr) >= RAM4 && RPAGE(addr) <= RAM7)
{
TSTATES += Machine->ulaDelay[TSTATES] & 0x7f;
}
#endif
#ifdef XZX_IF1
CHECK_AND_PAGEOUT_IF1(addr);
#endif
return op;
}
static uns8
spec128ReadMem(uns16 addr)
{
uns8 byte;
byte = MEM(addr);
#ifdef EMULATE_ULA_DELAY
if (RPAGE(addr) >= RAM4 && RPAGE(addr) <= RAM7)
{
TSTATES += Machine->ulaDelay[TSTATES] & 0x7f;
}
#endif
return byte;
}
static void
spec128WriteMem(uns16 addr, uns8 byte)
{
int page = VPAGE(addr);
if (RealPageNumbers[page] >= RAM0)
{
TheMemory[page][addr & 0x1fff] = byte;
if (RealPageNumbers[page] == ScreenSelect)
{
if ((addr &= 0x3fff) < (uns16)PIXEL_LENGTH)
{
ScreenWrite(addr, byte);
}
else if (addr < (uns16)(PIXEL_LENGTH + ATTR_LENGTH))
{
AttrWrite(addr, byte);
}
}
#ifdef EMULATE_ULA_DELAY
if (RPAGE(addr) >= RAM4 && RPAGE(addr) <= RAM7)
{
TSTATES += Machine->ulaDelay[TSTATES] & 0x7f;
}
#endif
}
}
static void
spec128InitPages(void)
{
Machine->pageIn(0, SP128ROM0);
Machine->pageIn(1, RAM5);
Machine->pageIn(2, RAM2);
Machine->pageIn(3, RAM0);
}
static void
spec128PageIn(int z80page, int vpage)
{
int i = z80page << 1;
TheMemory[i] = RealMemory[vpage];
RealPageNumbers[i] = vpage;
i++;
TheMemory[i] = RealMemory[vpage] + XZX_PAGESIZE / 2;
switch (vpage)
{
#ifdef XZX_MF
case MF128ROM:
RealPageNumbers[i] = MFRAM;
break;
#endif
#ifdef XZX_MGT
case MGTROM:
RealPageNumbers[i] = MGTRAM;
break;
#endif
default:
RealPageNumbers[i] = vpage;
break;
}
}
static inline void
spec128OutBankm128(uns8 val)
{
DEB(Msg(M_DEBUG,
"bankm: RAM select %d, ROM select %d, screen %d, paging %sabled",
val & B_SELRAM, (val & B_SELROM) >> 4, (val & B_SELSCREEN) >> 3,
PAGING_DISABLED(Last0x7FFD) ? "dis" : "en"););
if (Last0x7FFD == val || PAGING_DISABLED(Last0x7FFD))
{
return;
}
ScreenSelect = val & B_SELSCREEN ? RAM7 : RAM5;
if ((val & B_SELSCREEN) != (Last0x7FFD & B_SELSCREEN))
{
ForceScreenRefresh();
}
Last0x7FFD = val;
Machine->pageIn(3, RAM0 + (val & B_SELRAM));
#ifdef XZX_MF
if (RPAGE(0) != MF128ROM)
#endif
{
Machine->pageIn(0, SP128ROM0 + ((val & B_SELROM) >> 4));
}
}
static void
spec128OutPort(uns16 port, uns8 val)
{
DEB(Msg(M_DEBUG, "writing <%02x> to port <%04x>", val, port););
#ifdef XZX_TRDOS
OUTBYTE_TRDOS(port, val);
#endif
#ifdef XZX_MGT
OUTBYTE_MGT(port, val);
#endif
if ((port & B_ULA) == (P_ULA & B_ULA))
{
if (LastBorderColor != (val & B_BORDER))
{
LastBorderColor = val & B_BORDER;
SetBorderColor(LastBorderColor);
}
#ifdef SPEAKER_AUDIO
SpeakerOutByte(val);
#endif
#ifdef EMULATE_ULA_DELAY
TSTATES += Machine->ulaDelay[TSTATES] & 0x7f;
#endif
}
if ((port & B_128) == (P_128 & B_128))
{
if ((port & B_SNDCHIP) == (P_SNDCONTROL & B_SNDCHIP))
{
if ((port & B_SNDCONTROL) == (P_SNDCONTROL & B_SNDCONTROL))
{
Last0xFFFD = val;
}
else
{
if (Last0xFFFD < NELEM(Psg))
{
Psg[Last0xFFFD] = val;
if (Last0xFFFD == 14)
{
DEB(Msg(M_DEBUG, "writing AY register 14"););
}
#ifdef AYCHIP_AUDIO
else if (Last0xFFFD < 14)
{
AyOutByte(Last0xFFFD, val);
}
#endif
}
else
{
DEB(Msg(M_DEBUG, "writing AY register %d", Last0xFFFD););
}
}
}
else
{
spec128OutBankm128(val);
}
}
#ifdef XZX_IF1
OUTBYTE_IF1(port, val);
#endif
#ifdef PSEUDO_IO
if (port == PSEUDO_STDOUT)
{
(void)putc((int)val, stdout);
}
else if (port == PSEUDO_STDERR)
{
(void)putc((int)val, stderr);
}
#endif
OutPorts[LOWBYTE(port)] = val;
}
static uns8
spec128InPort(uns16 port)
{
uns8 res;
#ifdef XZX_TRDOS
INBYTE_TRDOS(port, res, quit);
#endif
#ifdef XZX_MGT
INBYTE_MGT(port, res, quit);
#endif
if ((port & B_KEMPSTON) == (P_KEMPSTON & B_KEMPSTON))
{
#ifdef JOY
if (GETCFG(joyActive) > 0)
{
InPorts[P_KEMPSTON] = (uns8)JoyRead();
}
#endif
res = InPorts[P_KEMPSTON] &
(B_RIGHT | B_LEFT | B_DOWN | B_UP | B_FIRE);
goto quit;
}
if ((port & B_ULA) == (P_ULA & B_ULA))
{
res = OutPorts[P_ULA] & B_SPEAKER ? 0xff : 0xbf;
if (!(port & 0x8000))
{
res &= KeyPorts[7];
}
if (!(port & 0x4000))
{
res &= KeyPorts[6];
}
if (!(port & 0x2000))
{
res &= KeyPorts[5];
}
if (!(port & 0x1000))
{
res &= KeyPorts[4];
}
if (!(port & 0x0800))
{
res &= KeyPorts[3];
}
if (!(port & 0x0400))
{
res &= KeyPorts[2];
}
if (!(port & 0x0200))
{
res &= KeyPorts[1];
}
if (!(port & 0x0100))
{
res &= KeyPorts[0];
}
res ^= TpGetNextBit();
#ifdef EMULATE_ULA_DELAY
TSTATES += 1;
#endif
goto quit;
}
if ((port & B_128) == (P_128 & B_128))
{
if ((port & B_SNDCHIP) == (P_SNDCONTROL & B_SNDCHIP))
{
if ((port & B_SNDCONTROL) == (P_SNDCONTROL & B_SNDCONTROL))
{
if (Last0xFFFD == 14)
{
DEB(Msg(M_DEBUG, "reading AY register 14"););
res = Psg[7] & 0x40 ? (Psg[14] & 0x1f) | 0x20 : 0x3f;
}
else if (Last0xFFFD < 14)
{
res = (Psg[Last0xFFFD] & PsgMask[Last0xFFFD]);
}
else if (Last0xFFFD == 15)
{
res = (Psg[7] & 0x80 ? Psg[15] : 0xff);
}
else
{
res = 0x00;
}
}
else
{
res = 0xff;
}
goto quit;
}
}
#ifdef XZX_IF1
INBYTE_IF1(port, res, quit);
#endif
#ifdef XZX_MF
INBYTE_MF128(port, res, quit);
#endif
#ifdef PSEUDO_IO
if (port == PSEUDO_STDIN)
{
res = (uns8)getchar();
goto quit;
}
#endif
if (!(Machine->ulaDelay[TSTATES % Machine->tstatesPerFrame] & 0x80))
{
unsigned int tstatesInLine = TSTATES % Machine->tstatesPerLine;
unsigned int line = Vline - Machine->topBorder;
res = RealMemory[ScreenSelect][(tstatesInLine & 0x02 ?
Line2Attr[line] : Line2Pix[line]) +
(tstatesInLine >> 2)];
}
else
{
res = 0xff;
}
quit:
DEB(Msg(M_DEBUG, "reading <%02x> from port <%04x>", res, port););
return res;
}
